/**
 * @license
 * Copyright 2017 Google Inc. All Rights Reserved.
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * =============================================================================
 */

import {keep, tidy} from '../../globals';
import {NDArrayMath} from '../../math';
import {SymbolicTensor} from '../graph';
import * as graph_util from '../graph_util';
import {SummedTensorArrayMap, TensorArrayMap} from '../tensor_array_map';
import {Operation} from './op';

/**
 * @hidden
 */
export class Log extends Operation {
  /**
   * Natural log operation - ln(x)
   */
  constructor(
      private xTensor: SymbolicTensor, private yTensor: SymbolicTensor) {
    super();
  }

  feedForward(math: NDArrayMath, inferenceArrays: TensorArrayMap) {
    const x = inferenceArrays.get(this.xTensor);

    tidy(() => {
      inferenceArrays.set(this.yTensor, keep(math.log(x)));
    });
  }

  backProp(
      math: NDArrayMath, inferenceArrays: TensorArrayMap,
      gradientArrays: SummedTensorArrayMap) {
    const x = inferenceArrays.get(this.xTensor);
    const dy = gradientArrays.get(this.yTensor);

    tidy(() => {
      if (graph_util.shouldBackProp(this.xTensor)) {
        gradientArrays.add(this.xTensor, math.divide(dy, x));
      }
    });
  }
}
